import logging

''' This class is used to validate the config
'''
class Schema:
    def __init__(self):
        pass

    def _validate(self):
        pass

    def __hyphen_to_underscore(self, k):
        return k.replace('-', '_')

    def _populate(self, data, definition):
        try:
            for (k, v) in data.items():
                k = self.__hyphen_to_underscore(k)
                if k in definition:
                    vtype = definition[k]
                    if isinstance(vtype, list):
                        assert isinstance(v, list), "{} should be a list".format(k)
                        newlist = []
                        for item in v:
                            newlist.append(vtype[0](item))
                        setattr(self, k, newlist)
                    else:
                        setattr(self, k, vtype(v))
            self._validate()
        except AssertionError as err:
            logging.error("Invalid config: %s", str(err))
            raise Exception("Invalid config")

    def _assert_non_empty(self, *attrs):
        for attr in attrs:
            assert len(getattr(self, attr)) > 0, "{}.{} must not be empty".format(type(self).__name__, attr)

    def _assert_non_null(self, *attrs):
        for attr in attrs:
            assert getattr(self, attr) is not None, "{}.{} must not be None".format(type(self).__name__, attr)


class InterceptQuerySettings(Schema):
    def __init__(self, data):
        self.plugin = None
        self.function = None

        self._populate(data, {
            'plugin': str,
            'function': str
        })

    def _validate(self):
        self._assert_non_null('plugin', 'function')
        self._assert_non_empty('plugin', 'function')


class InterceptCommandSettings(Schema):
    def __init__(self, data):
        self.queries = []
        self.connects = None

        self._populate(data, {
            'queries': [InterceptQuerySettings],
            'connects': str
        })


class InterceptSettings(Schema):
    def __init__(self, data):
        self.commands = None
        self.responses = None

        self._populate(data, {
            'commands': InterceptCommandSettings,
            'responses': str
        })


class Connection(Schema):
    def __init__(self, data):
        self.name = None
        self.host = None
        self.port = None

        self._populate(data, {
            'name': str,
            'host': str,
            'port': int
        })

    def _validate(self):
        self._assert_non_null('name', 'host', 'port')
        self._assert_non_empty('name')


class InstanceSettings(Schema):
    def __init__(self, data):
        self.listen = None
        self.redirect = None
        self.intercept = None

        self._populate(data, {
            'listen': Connection,
            'redirect': Connection,
            'intercept': InterceptSettings
        })


    def _validate(self):
        self._assert_non_null('listen', 'redirect')


class Settings(Schema):
    def __init__(self, data):
        self.log_level = None
        self.intercept_log = None
        self.general_log = None

        self._populate(data, {
            'log_level': str,
            'intercept_log': str,
            'general_log': str
        })

    def _validate(self):
        self._assert_non_null('log_level', 'intercept_log', 'general_log')
        self._assert_non_empty('log_level', 'intercept_log', 'general_log')


class Config(Schema):
    def __init__(self, data):
        self.plugins = []
        self.settings = None
        self.instances = []

        self._populate(data, {
            'plugins' : [str],
            'settings' : Settings,
            'instances' : [InstanceSettings]
        })

    def _validate(self):
        self._assert_non_empty('instances')



import logging

class Interceptor:
    def __init__(self, interceptor_config, plugins, context):
        self.interceptor_config = interceptor_config
        self.plugins = plugins
        self.context = context

    def intercept(self, packet_type, data):
        return data

    def get_codec(self):
        if self.context is not None and 'connect_params' in self.context:
            if self.context['connect_params'] is not None and 'client_encoding' in self.context['connect_params']:
                return self.context['connect_params']['client_encoding']
        return 'utf-8'


class CommandInterceptor(Interceptor):
    def intercept(self, packet_type, data):
        if self.interceptor_config.queries is not None:
            ic_queries = self.interceptor_config.queries
            if packet_type == b'Q':
                # Query, ends with b'\x00'
                data = self.__intercept_query(data, ic_queries)
            elif packet_type == b'P':
                # Statement that needs parsing.
                # First byte of the body is some Statement flag. Ignore, don't lose
                # Next is the query, same as above, ends with an b'\x00'
                # Last 2 bytes are the number of parameters. Ignore, don't lose
                statement = data[0:1]
                query = self.__intercept_query(data[1:-2], ic_queries)
                params = data[-2:]
                data = statement + query + params
            elif packet_type == b'':
                self.__intercept_context_data(data)
        return data


    def __intercept_context_data(self, data):
        # first 4 bytes and last zero byte are not interesting
        relevant_data = data[4:-1]
        # Each entry is terminated by b'\x00'
        entries = relevant_data.split(b'\x00')[:-1]
        entries = dict(zip(entries[0::2], entries[1::2]))
        self.context['connect_params'] = {}
        # Try to set codec, then transcode the dict
        if b'client_encoding' in entries:
            self.context['connect_params']['client_encoding'] = entries[b'client_encoding'].decode('ascii')
        codec = self.get_codec()
        for k, v in entries.items():
            self.context['connect_params'][k.decode(codec)] = v.decode(codec)


    def __intercept_query(self, query, interceptors):
        logging.getLogger('intercept').debug("intercepting query\n%s", query)
        # Remove zero byte at the end
        query = query[:-1].decode('utf-8')
        for interceptor in interceptors:
            if interceptor.plugin in self.plugins:
                plugin = self.plugins[interceptor.plugin]
                if hasattr(plugin, interceptor.function):
                    func = getattr(plugin, interceptor.function)
                    query = func(query, self.context)
                    logging.getLogger('intercept').debug(
                        "modifying query using interceptor %s.%s\n%s",
                        interceptor.plugin,
                        interceptor.function,
                        query)
                else:
                    raise Exception("Can't find function {} in plugin {}".format(
                        interceptor.function,
                        interceptor.plugin
                    ))
            else:
                raise Exception("Plugin {} not loaded".format(interceptor.plugin))
        # Append the zero byte at the end
        return query.encode('utf-8') + b'\x00'


class ResponseInterceptor(Interceptor):
    pass



'''For every configured instance, a Proxy object is created, that starts a listener.
On connect, it initiates a parallel connection to postgresql and pairs them together.
Using selectors, packets are received, intercepted and relayed to the other party.

Protocol:
The challenge is in identifying 3 types of packets:
1. With type and data.
   ex. 1 byte for type identifier, 4 bytes header for header and body length, body. Usually the body is ended with
   0x00 byte as well, that is part of the length.
   The queries are part of this type of packets. A query is b'Q####SELECT whatever\\x00'
2. Without type. They contain just a 4 byte header with packet length. It just so happens that the first byte is 0x00
   just because nothing is that long. These contain information about connection.
   Usually it's the client sending connection information. Ex.
        b'x00x00x00O' - length
        b'x00x03x00x00' - unexplained
        then, separated by x00 is a list of key, value: user, database, application_name, client_encoding, etc
        then, ended by b'x00'
3. Without data. Just the type. Since it's b'N', it might be "null"? The whole packet is this single byte.
   Signals "ok" according to wireshark

Handling:
proxy.py - connections and sockets things
connection.py - parsing and composing packets, launching interceptors
interceptors.py - intercepting for modification
'''

import logging
import selectors
import socket
import types
from postgresql_proxy import connection, config_schema as cfg
from postgresql_proxy.interceptors import ResponseInterceptor, CommandInterceptor


class Proxy(object):

    def __init__(self, instance_config, plugins):
        self.plugins = plugins
        self.num_clients = 0
        self.instance_config = instance_config
        self.connections = []
        self.selector = selectors.DefaultSelector()


    def __create_pg_connection(self, address, context):
        redirect_config = self.instance_config.redirect

        pg_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        pg_sock.connect((redirect_config.host, redirect_config.port))
        pg_sock.setblocking(False)

        events = selectors.EVENT_READ | selectors.EVENT_WRITE

        pg_conn = connection.Connection(pg_sock,
                                        name    = redirect_config.name + '_' + str(self.num_clients),
                                        address = address,
                                        events  = events,
                                        context = context)

        logging.info("initiated client connection to %s:%s called %s",
                     redirect_config.host, redirect_config.port, redirect_config.name)
        return pg_conn


    def __register_conn(self, conn):
        try:
            self.selector.register(conn.sock, conn.events, data=conn)
        except Exception:
            # potentially already registered - this can happen if file descriptors
            # are reused for new sockets -> try to unregister/re-register
            self.selector.unregister(conn.sock)
            self.selector.register(conn.sock, conn.events, data=conn)


    def __unregister_conn(self, conn):
        self.selector.unregister(conn.sock)


    def accept_wrapper(self, sock):
        clientsocket, address = sock.accept()  # Should be ready to
        clientsocket.setblocking(False)
        self.num_clients+=1
        sock_name = '{}_{}'.format(self.instance_config.listen.name, self.num_clients)
        logging.info("connection from %s, connection initiated %s", address, sock_name)

        events = selectors.EVENT_READ | selectors.EVENT_WRITE

        # Context dictionary, for sharing state data, connection details, which might be useful for interceptors
        context = {
            'instance_config': self.instance_config
        }

        conn = connection.Connection(clientsocket,
                                     name    = sock_name,
                                     address = address,
                                     events  = events,
                                     context = context)

        pg_conn = self.__create_pg_connection(address, context)

        if self.instance_config.intercept is not None and self.instance_config.intercept.responses is not None:
            pg_conn.interceptor = ResponseInterceptor(self.instance_config.intercept.responses, self.plugins, context)
            pg_conn.redirect_conn = conn

        if self.instance_config.intercept is not None and self.instance_config.intercept.commands is not None:
            conn.interceptor = CommandInterceptor(self.instance_config.intercept.commands, self.plugins, context)
            conn.redirect_conn = pg_conn

        self.__register_conn(conn)
        self.__register_conn(pg_conn)


    def service_connection(self, key, mask):
        sock = key.fileobj
        conn = key.data
        if mask & selectors.EVENT_READ:
            logging.debug('%s can receive', conn.name)
            recv_data = sock.recv(4096)  # Should be ready to read
            if recv_data:
                logging.debug('%s received data:\n%s', conn.name, recv_data)
                conn.received(recv_data)
            else:
                logging.info('%s connection closing %s', conn.name, conn.address)
                sock.close()
        if mask & selectors.EVENT_WRITE:
            if conn.out_bytes:
                logging.debug('sending to %s:\n%s', conn.name, conn.out_bytes)
                sent = sock.send(conn.out_bytes)  # Should be ready to write
                conn.sent(sent)



    def listen(self, max_connections = 8):
        '''Listen server socket. On connect launch a new thread with the client connection as an argument
        '''
        lconf = self.instance_config.listen
        ip, port = (lconf.host, lconf.port)
        try:
            logging.info("listening to %s:%s", ip, port)
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.bind((ip, port))
            self.sock.listen(max_connections)
            self.sock.setblocking(False)
            self.selector.register(self.sock, selectors.EVENT_READ, data=None)
            while True:
                events = self.selector.select(timeout=None)
                hit = False
                for key, mask in events:
                    hit = True
                    if key.data is None:
                        self.accept_wrapper(key.fileobj)
                    else:
                        self.service_connection(key, mask)
        except OSError as ex:
            logging.critical("Can't establish listener", exc_info=ex)
        finally:
            self.sock.close()
            self.sock = None
            logging.info("closed socket")


if(__name__=='__main__'):
    import importlib, yaml, os

    path = os.path.dirname(os.path.realpath(__file__))
    config = None
    try:
        with open(path + '/' + 'config.yml', 'r') as fp:
            config = cfg.Config(yaml.load(fp))
    except Exception:
        logging.critical("Could not read config. Aborting.")
        exit(1)

    logging.basicConfig(
        filename=config.settings.general_log,
        level=getattr(logging, config.settings.log_level.upper()),
        format='%(asctime)s : %(levelname)s : %(message)s'
    )

    qlog = logging.getLogger('intercept')
    qformat = logging.Formatter('%(asctime)s : %(message)s')
    qhandler = logging.FileHandler(config.settings.intercept_log, mode = 'w')
    qhandler.setFormatter(qformat)
    qlog.addHandler(qhandler)
    qlog.setLevel(logging.DEBUG)

    print('general log, level {}: {}'.format(config.settings.log_level, config.settings.general_log))
    print('intercept log: {}'.format(config.settings.intercept_log))
    print('further messages directed to log')

    plugins = {}
    for plugin in config.plugins:
        logging.info("Loading module %s", plugin)
        module = importlib.import_module('plugins.' + plugin)
        plugins[plugin] = module

    for instance in config.instances:
        logging.info("Starting proxy instance")
        proxy = Proxy(instance, plugins)
        proxy.listen()



import os
import re
import sys
from setuptools import find_packages, setup

install_requires = []


if __name__ == '__main__':

    setup(
        name='postgresql-proxy',
        version='0.0.2',
        description='Postgresql Proxy',
        packages=find_packages(exclude=('tests', 'tests.*')),
        install_requires=install_requires,
        zip_safe=False,
        classifiers=[
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'License :: OSI Approved :: Apache Software License',
            'Topic :: Software Development :: Testing',
        ]
    )



import logging

class Connection:
    def __init__(self, sock, address, name, events, context):
        self.sock = sock
        self.address = address
        self.name = name
        self.events = events
        self.context = context
        self.interceptor = None
        self.redirect_conn = None
        self.out_bytes = b''
        self.in_bytes = b''

    def parse_length(self, length_bytes):
        return int.from_bytes(length_bytes, 'big')

    def encode_length(self, length):
        return length.to_bytes(4, byteorder='big')

    def received(self, in_bytes):
        self.in_bytes += in_bytes
        # Read packet from byte array while there are enough bytes to make up a packet.
        # Otherwise wait for more bytes to be received (break and exit)
        while True:
            ptype = self.in_bytes[0:1]
            if ptype == b'\x00':
                if len(self.in_bytes) < 4:
                    break
                header_length = 4
                body_length = self.parse_length(self.in_bytes[0:4]) - 4
            elif ptype == b'N':
                header_length = 1
                body_length = 0
            else:
                if len(self.in_bytes) < 5:
                    break
                header_length = 5
                body_length = self.parse_length(self.in_bytes[1:5]) - 4

            length = header_length + body_length
            if len(self.in_bytes) < length:
                break
            header = self.in_bytes[0:header_length]
            body = self.in_bytes[header_length:length]
            self.process_inbound_packet(header, body)
            self.in_bytes = self.in_bytes[length:]

    def process_inbound_packet(self, header, body):
        if header != b'N':
            packet_type = header[0:-4]
            logging.info("intercepting packet of type '%s' from %s", packet_type, self.name)
            body = self.interceptor.intercept(packet_type, body)
            header = packet_type + self.encode_length(len(body) + 4)
        message = header + body
        logging.debug("Received message. Relaying. Speaker: %s, message:\n%s", self.name, message)
        self.redirect_conn.out_bytes += message

    def sent(self, num_bytes):
        self.out_bytes = self.out_bytes[num_bytes:]
